//////////////////////////////////////////////////////////////////////
//
// File:     DocSet.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     5/25/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.*;
import java.text.*;

// Note: Need to make this more flexible... allow some command
// line or other method for just checking single files.

// Main class - Represents a set of tokenized html documents
public class DocSet
	implements HtmlCollector
{
	/** Constants **/
	public final static int    _MAX_INPUT = 65536;
	public final static String _ANCHOR    = "<A>";
	public final static String _TD        = "<TD>";
	public final static String _TR        = "<TR>";
	public final static String _GENHEAD   = "<H?>";
	public final static String _GENCHEAD  = "</H?>";
	public final static DecimalFormat _df1 = new DecimalFormat("#0000");	
	public final static DecimalFormat _df2 = new DecimalFormat("#00.0");	
	public final static DecimalFormat _df3 = new DecimalFormat("#0.000");	

	/** Member Fields **/
	public String    _sFilename;    // The filename which this set refers to...
	public ArrayList _alDocs;       // ArrayList of ArrayList of tokens
	public ArrayList _alDocNames;   // ArrayList of String for each doc name
	public ArrayList _alDocTargets; // ArrayList of ArrayList of target
									// indexes
	public ArrayList _alCurDoc;     // Current document index being
									// processed...
	public Template  _tTemplate;    // Contains parameters for hmm
									// construction

	public SuffStatistics _ssStat;  // Used to update sufficient
									// statistics for this document

	public int       _nCurState;    // Used during HtmlCollector callback,
	                                // determines if previous tag was a label

	/** Member Methods **/
	public DocSet(String setname, Template temp, SuffStatistics s)
	{
		_alDocs        = new ArrayList();
		_alDocNames    = new ArrayList();
		_alDocTargets  = new ArrayList();
		_sFilename     = setname;
		_tTemplate     = temp;
		_ssStat        = s;

		// For every string in file list, process the document,
		// including accumulating sufficient statistics
		loadSetNames(setname);
		processSetNames();
	}

	////////////////////////////////////////////////////////////
	//                Document Setup Routines
	////////////////////////////////////////////////////////////
	protected void loadSetNames(String setname)
	{
		_alDocNames.clear();
        if (setname == null) return;
		
        try {
            String varname;
            InputStream in = new FileInputStream(setname);
            byte[] b = new byte[_MAX_INPUT];
            
            int l = in.read(b);
            String s = new String(b);
            StringTokenizer st = new StringTokenizer(s.substring(0,l-1));
            
            while (st.hasMoreTokens())
                {
                    varname = st.nextToken();
					_alDocNames.add(varname);
                    //if (!st.hasMoreTokens()) break;
                    //value = st.nextToken();
                    //setBindings(varname, value);
                }
        } catch (IOException e) 
			{
			System.out.println("\nError while processing '" + setname + "', skipping...");
		}
	}

    protected void processSetNames()
	{
		int i, num_docs = getNumDocuments();

		// Initialize all data structures to be filled
		System.out.println("*Processing '" + _sFilename + "'...");
		_alDocs.clear();
		_alDocTargets.clear();
		if (_ssStat != null) _ssStat.resetSufficientStats();

		for (i=0; i<num_docs; i++)
		{
			_alDocs.add(new ArrayList());
			_alDocTargets.add(new ArrayList());
			ProcessFile(i);
		}
	}

	protected void ProcessFile(int index)
	{
		String docname = (String)_alDocNames.get(index); 
		System.out.println("Processing '" + docname + "'...");

		// Process the document tokens
		_alCurDoc = (ArrayList)_alDocs.get(index);
		_alCurDoc.add(new DocToken("[START]", DocToken._SDOC, DocToken._START));
		_nCurState = DocToken._UNKNOWN;
		HtmlExtractor ex = new HtmlExtractor();
		ex.extract(docname, new HtmlExporter(this));
		_alCurDoc.add(new DocToken("[END]", DocToken._EDOC, DocToken._END)); 

		// Now post-process the target to initialize the
		// suffix/prefix
		setDocTargets(index);
	}
	
	////////////////////////////////////////////////////////////
	//                Document Access Interface
	////////////////////////////////////////////////////////////
	public SuffStatistics getSufficientStats()
	{
		return _ssStat;
	}
	
	public String getDocName(int index)
	{
		return (String)_alDocNames.get(index);
	}
	
	public int getNumDocuments()
	{
		return _alDocNames.size();
	}

	public ArrayList getDocument(int index)
	{
		return (ArrayList)_alDocs.get(index);
	}

	public int getDocLength(int index)
	{
		return ((ArrayList)_alDocs.get(index)).size();
	}

	public DocToken getDocumentToken(int index, int tok)
	{
		return (DocToken)((ArrayList)_alDocs.get(index)).get(tok);
	}

	public int getNumDocTargets(int index)
	{
		return ((ArrayList)_alDocTargets.get(index)).size();
	}

	public int getDocTargetIndex(int index, int ntarget)
	{
		return ((Integer)((ArrayList)_alDocTargets.get(index)).get(ntarget)).intValue();
	}

	public void addDocTarget(int index, int tindex)
	{
		((ArrayList)_alDocTargets.get(index)).add(new Integer(tindex));
	}

	//////////////////////////////////////////////////////////////
	//  Document Token State/Sufficient Statistic Initialization
	//////////////////////////////////////////////////////////////
	public void resetDocTargets(int index)
    {
		int e = getDocLength(index) - 1;
		ArrayList doc = getDocument(index);

		if ( (((DocToken)doc.get(0))._nClass != DocToken._START) ||
			 (((DocToken)doc.get(e))._nClass != DocToken._END) )
		{
			System.out.println("Error identifying start/end tokens during reset, halting...");
			System.exit(1);
		}	 
		//((DocToken)doc.get(0))._nClass = DocToken._START;
		//if (e > 0) ((DocToken)doc.get(e))._nClass = DocToken._END;

		for (int i=1; i<e; i++)
			((DocToken)doc.get(i))._nClass = DocToken._UNKNOWN;
	}

    public void setDocTargets(int index)
    {
		DocToken  dt;
		ArrayList toklist = (ArrayList)_alDocTargets.get(index);;
		ArrayList doc     = getDocument(index);
		int       e       = getDocLength(index) - 1;
		int       pre_start, suf_end, i, j;

		// If the context width is < zero, this is a document to process
		if (_tTemplate._nContextWidth == Template._TREAT_AS_UNLABELED) 
		{
			resetDocTargets(index);
			return;
		}

		if ( (((DocToken)doc.get(0))._nClass != DocToken._START) ||
			 (((DocToken)doc.get(e))._nClass != DocToken._END) )
		{
			System.out.println("Error identifying start/end tokens during processing, halting...");
			System.exit(1);
		}	 
		//((DocToken)doc.get(0))._nClass = DocToken._START;
		//if (e > 0) ((DocToken)doc.get(e))._nClass = DocToken._END;

		// Look for target indices while setting all others
		// to background
		toklist.clear();
		for (i=1; i<e; i++)
			{
				dt = (DocToken)doc.get(i);
				if (dt._nClass != DocToken._TARGET)
					dt._nClass = DocToken._BKGD;
				else
					toklist.add(new Integer(i));
			}

		// Now process prefixes and suffixes for the window size
		Iterator it = toklist.iterator();
		while (it.hasNext())
		{
			i = ((Integer)it.next()).intValue();
			pre_start = i - _tTemplate._nContextWidth;
			suf_end   = i + _tTemplate._nContextWidth;
			if (pre_start <= 0) pre_start = 1;
			if (suf_end   >= e) suf_end   = e-1;
			
			// Process prefixes
			for (j=pre_start; j < i; j++)
			{
				dt = (DocToken)doc.get(j);
				if (dt._nClass != DocToken._TARGET)
				{
					if ((dt._nClass == DocToken._SUFFIX) ||
						(dt._nClass == DocToken._BOTH))
					{							
						//System.out.println("Labeling " + j + ":" + 
						//				   dt._sWord + " as BOTH");
						
						dt._nClass = DocToken._BOTH;
					}
					else 
					{
						//System.out.println("Labeling " + j + ":" + 
						//				   dt._sWord + " as PREFIX");
						dt._nClass = DocToken._PREFIX;
					}
				}
			}  
			
			// Process suffixes
			for (j=i+1; j <= suf_end; j++)
			{
				dt = (DocToken)doc.get(j);
				if (dt._nClass != DocToken._TARGET)
				{
					//System.out.println("Labeling " + j + ":" + 
					//				   dt._sWord + " as SUFFIX");
					dt._nClass = DocToken._SUFFIX;
				}
			}  
		}

		// Process this document for sufficient statistics
		if (_ssStat != null) 
			processSufficientStats(index);
	}
	
	public void processSufficientStats(int index)
	{
		int       pos = 0;
		int       len = getDocLength(index); 
		ArrayList doc = getDocument(index);
		DocToken  cur, prev;

		// We do have an initial observation and state
		System.out.println("Processing suff stats for '" + getDocName(index) + "'...");
		prev = (DocToken)doc.get(pos++);
		_ssStat.incObsPriors(prev._nObsIndex);
		_ssStat.incStatePriors(prev._nClass);
		_ssStat.AddObsFromState(prev._nObsIndex, prev._nClass);

		while (pos < len)
		{
			cur = (DocToken)doc.get(pos++);

			// Update state transition probability
			_ssStat.AddStateFromState(cur._nClass, prev._nClass);

			// Update obs transition probability
			_ssStat.AddObsFromState(cur._nObsIndex, cur._nClass);

			// Update state priors
			_ssStat.incStatePriors(cur._nClass);

			// Update observation priors
			_ssStat.incObsPriors(cur._nObsIndex);
			
			// Next...
			prev = cur;
		}
	}

	////////////////////////////////////////////////////////////
	//      Implementation of the HtmlCollector interface.
	////////////////////////////////////////////////////////////
	public int isTarget(String s)
	{
		return (s.toUpperCase().startsWith("<!--T-" + _tTemplate._sTargetName))
			? DocToken._TARGET : DocToken._UNKNOWN;
	}

	public void finish()
	{
		// Do nothing for now...
	}

	public void addTag(String t)
	{
		int cur_state = _nCurState;
		String ct = t.toUpperCase();
		_nCurState = DocToken._UNKNOWN;

		if (_tTemplate._bIgnoreTags)
			return;

		if (ct.startsWith("<A "))
			_alCurDoc.add(new DocToken(_ANCHOR, DocToken._TAG, cur_state, t));
		else if (ct.startsWith("<TD "))
			_alCurDoc.add(new DocToken(_TD, DocToken._TAG, cur_state, t));
		else if (ct.startsWith("<TR "))
			_alCurDoc.add(new DocToken(_TR, DocToken._TAG, cur_state, t));
		else if (ct.startsWith("<H"))
			_alCurDoc.add(new DocToken(_GENHEAD, DocToken._TAG, cur_state, t));
		else
			_alCurDoc.add(new DocToken(ct, DocToken._TAG, cur_state, t));
	}

	public void addCloseTag(String t)
	{
		int cur_state = _nCurState;
		String ct = t.toUpperCase();
		_nCurState = DocToken._UNKNOWN;

		if (_tTemplate._bIgnoreTags)
			return;

		if (ct.startsWith("</H"))
			_alCurDoc.add(new DocToken(_GENCHEAD, DocToken._TAG, cur_state, t));
		else
			_alCurDoc.add(new DocToken(ct, DocToken._TAG, cur_state, t));
	}

	public void addComment(String t)
	{
		// Generally discarded except when marking special states
		_nCurState = isTarget(t);
	}

	public void addNewline(String t)
	{
		// Discard altogether...
	}

	public void addAnnotation(String t)
	{
		// I have no idea what these are, discard...
	}

	public void addTextString(String t)
	{
		int cur_state = _nCurState;
		_nCurState = DocToken._UNKNOWN;

		// Tokenize the text string
		DocToken dt;
		TextTokenizer tt = new TextTokenizer(t);
		while ((dt = tt.nextToken(cur_state)) != null)
			_alCurDoc.add(dt);

		//new DocToken(ct, DocToken._TAG, cur_state, t));
	}

	////////////////////////////////////////////////////////////
	//                 Document I/O Utilities
	////////////////////////////////////////////////////////////
	public void exportResults(PrintStream ps)
	{
		int i, ndocs = getNumDocuments();
		boolean prev_target;

		ps.println("Extraction results for document set '" + _sFilename + "':\n"
				   +"========================================================\n");
		for(i=0; i<ndocs; i++)
		{
			ps.println("Extraction results for document:\n'" + 
					   getDocName(i) + "'" + 
					   "\n--------------------------"+
					   "----------------------------");
			
			// Go through the document, printing each token on a line...
			Iterator it = ((ArrayList)_alDocs.get(i)).iterator();
			DocToken dt;
			
			prev_target = false;
			while (it.hasNext())
			{
				dt = (DocToken)it.next();
				if (dt._nClass == DocToken._TARGET) 
				{
					if (!prev_target)
						ps.print(_tTemplate._sTargetName + ": ");
					ps.print(dt._sExtra + " ");
					prev_target = true;
				} 
				else 
				{
					if (prev_target)
						ps.println();
					prev_target = false;
				}
			}
			ps.println();
		}
	}

	public void printAllDocuments(PrintStream ps)
	{
		int max = getNumDocuments();
		for (int i=0; i<max; i++)
		{
			printDocument(i, ps);
			ps.println();
		}
	}
 
	public void printDocument(int index, PrintStream ps)
	{
		// Print header
		ps.println("File: " + getDocName(index) + 
						   "\n------------------\n");

		// Go through the document, printing each token on a line...
		Iterator i = ((ArrayList)_alDocs.get(index)).iterator();
		DocToken dt;
		
		while (i.hasNext())
		{
			dt = (DocToken)i.next();
			ps.println(dt);
		}
	}

	// Compare the two doc sets and report relevant statistics
	// Note: [this] is the answer set
	public void printComparisonResults(DocSet test, 
									   PrintStream ps)
	{		
		double ATAR_TTAR = 0.0; // Agreement
		double AOTH_TTAR = 0.0; // False positive
		double ATAR_TOTH = 0.0; // False negative
		double AOTH_TOTH = 0.0; // Agreement
		DocToken atoken, ttoken;
		double prec, recall, f1measure;

		// Traverse each document in pairs, updating the
		// counts incrementally
		ps.println("\nValidation Data Statistics:\n" +
				   "===========================\n");
		int doc, ndocs = getNumDocuments();
		int word, nwords;
		for (doc = 0; doc < ndocs; doc++)
		{
			nwords = getDocLength(doc);
			for (word = 0; word < nwords; word++)
			{
				atoken = getDocumentToken(doc, word);
				ttoken = test.getDocumentToken(doc, word);

				// Make sure to disregard non-words/numbers...
				if ((atoken._nType != DocToken._WORD) &&
					(atoken._nType != DocToken._NUM))
					continue;

				// Increment counts
				if (atoken._nClass == DocToken._TARGET)
				{
					if (ttoken._nClass == DocToken._TARGET)
						ATAR_TTAR++;
					else
						ATAR_TOTH++;
				}
				else
				{
					if (ttoken._nClass == DocToken._TARGET)
						AOTH_TTAR++;
					else
						AOTH_TOTH++;
				}
			}
		}

		// Compute statistics
		prec      = (ATAR_TTAR / (ATAR_TTAR + AOTH_TTAR));
		recall    = (ATAR_TTAR / (ATAR_TTAR + ATAR_TOTH));
		f1measure = (2*prec*recall) / (prec + recall);

		ps.println("           ANSWER");
		//ps.println();
		ps.println("         TAR    OTH");
		ps.println("       +------------");
		ps.println(" T  TAR| " + _df1.format(ATAR_TTAR) + "   " +
				   _df1.format(AOTH_TTAR));
		ps.println(" S     |");
		ps.println(" T  OTH| " + _df1.format(ATAR_TOTH) + "   " +
				   _df1.format(AOTH_TOTH) + "\n");
		
		ps.println(" Precision : " + _df2.format(100.0*prec) + "%");
		ps.println(" Recall    : " + _df2.format(100.0*recall) + "%");
		ps.println(" F1 Measure: " + _df3.format(f1measure));
		ps.println();
	}

}
