//////////////////////////////////////////////////////////////////////
//
// File:     HMM.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     4/12/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.util.*;
import java.lang.*;
import java.text.*;

// Note: Need to clean this up, make state collapsing more
// generic and allow for model export/import

// Main class - Hidden Markov Model (HMM) representation
public class HMM
{
	/** Constants **/
	public final static int _UNIFORM   = 0;
	public final static int _CONTEXT   = 1;
	public final static int _ALL       = 2;
	public final static int _NUM_TYPES = 3;
	public final static DecimalFormat _df = new DecimalFormat("#0.000");
	
	/** Members **/
	public int         _nType;        // The type of HMM
	public boolean[][] _a2LegalTrans; // Indicates legal transitions (in
									  // mapped space, not unique space)
	                                  // [to_state][from_state]
	public double[][]  _a2Trans;      // [to_state][from_state] cond prob
	public double[][]  _a2Obs;        // [obs][state] cond prob
	public int[]       _aStateMap;    // Map from given states to HMM states

	public SuffStatistics _ssStat;   

	/** Methods **/
	public HMM(int type)
	{
		if ((type < 0) || (type >= _NUM_TYPES))
		{
			System.out.println("Illegal HMM specification, halting...");
			System.exit(1);
		}

		_nType   = type;
		_ssStat  = null;

		// Allocate trans and obs arrays
		_a2Trans      = new double[DocToken._NUM_VALID_STATES][DocToken._NUM_VALID_STATES];
		_a2Obs        = new double[SuffStatistics._MAX_WORDS][DocToken._NUM_VALID_STATES];
		_a2LegalTrans = new boolean[DocToken._NUM_VALID_STATES][DocToken._NUM_VALID_STATES];
		_aStateMap    = new int[DocToken._NUM_STATES];
		
		// Initialize all arrays
		initialize();
	}

	public void initialize()
	{
		int i, j;

		// Initialize observation probabilities
		for (i=0; i<SuffStatistics._MAX_WORDS; i++)
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				_a2Obs[i][j] = 0.0;

		// Initialize transition probabilities
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				_a2Trans[i][j] = 0.0;

		// Initialize transitions... in the mapped space!
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				_a2LegalTrans[i][j] = false;

		// Based on the HMM type, initialize the state map and
		// transition map -- these maps are hardcoded for now
		switch (_nType)
		{
		case _UNIFORM: 
			{
				// All states are the same
				for (i=0; i<DocToken._NUM_VALID_STATES; i++)
					_aStateMap[i] = 0;

				// Note: This is in the mapped space!
				_a2LegalTrans[0][0] = true; // Only one legal trans!
			} break;
		case _CONTEXT:
			{
				// Suffix and Both collapse to Prefix
				for (i=0; i<DocToken._NUM_VALID_STATES; i++)
					_aStateMap[i] = i;
				_aStateMap[DocToken._SUFFIX] = DocToken._PREFIX;
				_aStateMap[DocToken._BOTH]   = DocToken._SUFFIX;

				// Note: This is in the mapped space!
				_a2LegalTrans[DocToken._BKGD][DocToken._START] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._START] = true;
				_a2LegalTrans[DocToken._END][DocToken._START] = true;
				_a2LegalTrans[DocToken._END][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._BKGD][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._TARGET][DocToken._TARGET] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._BKGD][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._END][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._TARGET][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._TARGET] = true;

			} break;
		case _ALL:
			{
				// All unique but make sure to catch the _BOTH case
				// during training/retrieval
				for (i=0; i<DocToken._NUM_VALID_STATES; i++)
					_aStateMap[i] = i;

				// Note: This is in the mapped space!
				_a2LegalTrans[DocToken._BKGD][DocToken._START] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._START] = true;
				_a2LegalTrans[DocToken._END][DocToken._START] = true;
				_a2LegalTrans[DocToken._BKGD][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._SUFFIX][DocToken._SUFFIX] = true;
				_a2LegalTrans[DocToken._TARGET][DocToken._TARGET] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._END][DocToken._BKGD] = true;
				_a2LegalTrans[DocToken._TARGET][DocToken._PREFIX] = true;
				_a2LegalTrans[DocToken._SUFFIX][DocToken._TARGET] = true;
				_a2LegalTrans[DocToken._PREFIX][DocToken._SUFFIX] = true;
				_a2LegalTrans[DocToken._BKGD][DocToken._SUFFIX] = true;
				_a2LegalTrans[DocToken._END][DocToken._SUFFIX] = true;

			} break;
		}
	}

	public void train(SuffStatistics s)
	{
		// (*)This is hardcoded for now, should be generalized once
		//    state collapsing is given a more well-defined notion...
		//    if two states collapsed, need to sum numerators and 
		//    denominators
		int i, j;
		_ssStat = s; // Most recently trained upon suff. statistics

		if (_nType == _UNIFORM)
		{
			// Only legal state here is zero, everything
			// is collapsed down to these states

			// Transition probability!
			double uniform_obs = 1.0/(double)_ssStat._nWords;

			// Make it look like equal transitions between all states
			_a2Trans[0][0] = 1.0/*/(double)DocToken._NUM_VALID_STATES*/;

			// Uniform observation probabilities
			for (i=0; i<_ssStat._nWords;i++)
				_a2Obs[i][0] = uniform_obs;
		}
		else if (_nType == _CONTEXT)
		{
			// See note(*) above for reason for doing this...

			// Calculate transition probabilties
			_a2Trans[DocToken._BKGD][DocToken._START] = 
					(double)_ssStat._a2StateFromState[DocToken._BKGD][DocToken._START] /
					(double)_ssStat._aStatePriors[DocToken._START];				
			_a2Trans[DocToken._PREFIX][DocToken._START] = 
					((double)_ssStat._a2StateFromState[DocToken._PREFIX][DocToken._START]+
					 (double)_ssStat._a2StateFromState[DocToken._SUFFIX][DocToken._START])/
					(double)_ssStat._aStatePriors[DocToken._START];				
			_a2Trans[DocToken._END][DocToken._START] = 
					(double)_ssStat._a2StateFromState[DocToken._END][DocToken._START] /
					(double)_ssStat._aStatePriors[DocToken._START];				
			_a2Trans[DocToken._END][DocToken._PREFIX] = 
					((double)_ssStat._a2StateFromState[DocToken._END][DocToken._PREFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._END][DocToken._SUFFIX]) /
				    ((double)_ssStat._aStatePriors[DocToken._PREFIX] + 
					 (double)_ssStat._aStatePriors[DocToken._SUFFIX]);				
			_a2Trans[DocToken._BKGD][DocToken._BKGD] = 
					(double)_ssStat._a2StateFromState[DocToken._BKGD][DocToken._BKGD] /
					(double)_ssStat._aStatePriors[DocToken._BKGD];				
			_a2Trans[DocToken._PREFIX][DocToken._PREFIX] = 
					((double)_ssStat._a2StateFromState[DocToken._PREFIX][DocToken._PREFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._PREFIX][DocToken._SUFFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._SUFFIX][DocToken._PREFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._SUFFIX][DocToken._SUFFIX]) /
				    ((double)_ssStat._aStatePriors[DocToken._PREFIX] + 
					 (double)_ssStat._aStatePriors[DocToken._SUFFIX]);				
			_a2Trans[DocToken._TARGET][DocToken._TARGET] = 
					(double)_ssStat._a2StateFromState[DocToken._TARGET][DocToken._TARGET] /
					(double)_ssStat._aStatePriors[DocToken._TARGET];				
			_a2Trans[DocToken._PREFIX][DocToken._BKGD] = 
					((double)_ssStat._a2StateFromState[DocToken._PREFIX][DocToken._BKGD] +
					 (double)_ssStat._a2StateFromState[DocToken._SUFFIX][DocToken._BKGD]) /
					(double)_ssStat._aStatePriors[DocToken._BKGD];				
			_a2Trans[DocToken._BKGD][DocToken._PREFIX] = 
					((double)_ssStat._a2StateFromState[DocToken._BKGD][DocToken._PREFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._BKGD][DocToken._SUFFIX]) /
				    ((double)_ssStat._aStatePriors[DocToken._PREFIX] + 
					 (double)_ssStat._aStatePriors[DocToken._SUFFIX]);				
			_a2Trans[DocToken._END][DocToken._BKGD] = 
					(double)_ssStat._a2StateFromState[DocToken._END][DocToken._BKGD] /
					(double)_ssStat._aStatePriors[DocToken._BKGD];				
			_a2Trans[DocToken._TARGET][DocToken._PREFIX] = 
					((double)_ssStat._a2StateFromState[DocToken._TARGET][DocToken._PREFIX] +
					 (double)_ssStat._a2StateFromState[DocToken._TARGET][DocToken._SUFFIX]) /
				    ((double)_ssStat._aStatePriors[DocToken._PREFIX] + 
					 (double)_ssStat._aStatePriors[DocToken._SUFFIX]);				
			_a2Trans[DocToken._PREFIX][DocToken._TARGET] = 
					((double)_ssStat._a2StateFromState[DocToken._PREFIX][DocToken._TARGET] + 
					 (double)_ssStat._a2StateFromState[DocToken._SUFFIX][DocToken._TARGET]) /
					(double)_ssStat._aStatePriors[DocToken._TARGET];				

			// Calculate observation probabilities
			for (i=0; i<_ssStat._nWords; i++)
			{
				_a2Obs[i][DocToken._START] = 
					(double)_ssStat._a2ObsFromState[i][DocToken._START] /
					(double)_ssStat._aStatePriors[DocToken._START];

				_a2Obs[i][DocToken._END] = 
					(double)_ssStat._a2ObsFromState[i][DocToken._END] /
					(double)_ssStat._aStatePriors[DocToken._END];

				_a2Obs[i][DocToken._PREFIX] = 
					((double)_ssStat._a2ObsFromState[i][DocToken._PREFIX] + 
					 (double)_ssStat._a2ObsFromState[i][DocToken._SUFFIX]) /
					((double)_ssStat._aStatePriors[DocToken._PREFIX] +
					 (double)_ssStat._aStatePriors[DocToken._SUFFIX]);

				_a2Obs[i][DocToken._TARGET] = 
					(double)_ssStat._a2ObsFromState[i][DocToken._TARGET] /
					(double)_ssStat._aStatePriors[DocToken._TARGET];

				_a2Obs[i][DocToken._BKGD] = 
					(double)_ssStat._a2ObsFromState[i][DocToken._BKGD] /
					(double)_ssStat._aStatePriors[DocToken._BKGD];
			}
		}
		else if (_nType == _ALL)
		{
			// Calculate transition probabilities...
			for (i=0; i<DocToken._NUM_VALID_STATES; i++)
				for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				{
					// We know in this case that there is a bijective
					// mapping between requested and actual states
					if (_a2LegalTrans[i][j])
					{
						_a2Trans[i][j] = 
							(double)_ssStat._a2StateFromState[i][j] /
							(double)_ssStat._aStatePriors[j];
						//System.out.println(_ssStat._a2StateFromState[i][j]
						//				   + " / " +
						//				   _ssStat._aStatePriors[j] +
						//				   " = " + _a2Trans[i][j]);
					}
				}

			// Calculate observation probabilities
			for (i=0; i<_ssStat._nWords; i++)
				for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				{
					if (_ssStat._aStatePriors[j] <= 0.0)
						continue;
					_a2Obs[i][j] = 
						(double)_ssStat._a2ObsFromState[i][j] /
						(double)_ssStat._aStatePriors[j];
				}
		}
	}

	// Compute the mixture transition probabilities
	public double getTransProb(int to_state, int from_state)
	{
		int to_map   = _aStateMap[to_state];
		int from_map = _aStateMap[from_state];

		// Collapsing has to preserve all links so this is OK
		if (!_a2LegalTrans[to_map][from_map])
		{
			return 0.0;
			//System.out.println("Illegal transition request, exiting...");
			//System.exit(1);
		}

		return _a2Trans[to_map][from_map];
	}

	// Compute the mixture observation probabilities
	public double getObsProb(int obs, int state)
	{
		int state_map = _aStateMap[state];

		if (obs == -1) 
		{
			if (_nType == _UNIFORM)
				return _a2Obs[0][state_map];
			else
				return 0.0;
		}

		// Collapsing has to preserve all links so this is OK
		if (obs >= _ssStat._nWords)
		{
			System.out.println("Illegal observation request, exiting...");
			System.exit(1);
		}

		return _a2Obs[obs][state_map];
	}

	// Print out the transition and observation probabilities
	public void printProbabilities(PrintStream ps)
	{
		int i, j;
		// Calculate transition probabilities...
		ps.println("\nTransition Probabilities: row(to), col(from)\n" +
				   "--------------------------------------------");
		ps.println("          STA    BKG    PRE    SUF    TAR    END");
		ps.println("         -----  -----  -----  -----  -----  -----");
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
		{
			ps.print(DocToken.getClassificationName(i) + ": ");
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
		    {
				ps.print(_df.format(_a2Trans[i][j]) + "  ");
			}
			ps.println();
		}

		// Calculate observation probabilities
		ps.println("\nObservation Probabilities: row(obs), col(state)\n" +
				   "-----------------------------------------------");
		ps.println("          STA    BKG    PRE    SUF    TAR    END");
		ps.println("         -----  -----  -----  -----  -----  -----");
		for (i=0; i<_ssStat._nWords; i++)
		{
			ps.print(formatWord(i) + ": ");
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
			{
				ps.print(_df.format(_a2Obs[i][j]) + "  ");
			}
			ps.println();
		}
		ps.println();
	}

	// Returns an observation word of length 7
	public String formatWord(int index)
	{
		int i, len;
		String ret = _ssStat.getWord(index);
		if (ret.length() > 7)
			ret = ret.substring(0,7);
		len = ret.length();
		for (i=len; i<7; i++)
			ret += " ";
		return ret;
	}
}
