//////////////////////////////////////////////////////////////////////
//
// File:     MixtureHMM.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     4/12/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.util.*;
import java.lang.*;
import java.text.*;

// Note 1: It seems that there should be a layer between this
//         HMM mixture model and the document interface, this
//         would allow the HMM, MixtureHMM, and SuffStatistics
//         to be abstracted from this task.  For the time being
//         however, this abstraction is not made.
//
//         This should also be abstracted to have the same
//         interface as the generic HMM.

// Note 2: Need to clean this up, make state collapsing more
// generic and allow for model export/import (and recursive
// sub-model import/export).

// Note 3: What about adaptive weighting based on 'experience'?


// Main class - Implements a mixture model of HMM's
public class MixtureHMM
{
	/** Constants **/
	public final static double _PINF =  Double.MAX_VALUE;
	public final static double _NINF = -Double.MAX_VALUE;

	/** Members **/
	public ArrayList      _alHMM;    // List of HMM's
	public ArrayList      _alLambda; // List of weights for each HMM
	public SuffStatistics _ssStat;   // Object which contains suff stats,
	                                 // All Mixture HMM's have to share same
	                                 // sufficient statistics
	public final static DecimalFormat _df = new DecimalFormat("#0.00000");

	/** Methods **/
	public MixtureHMM()
	{
		_alHMM    = new ArrayList();
		_alLambda = new ArrayList();
		_ssStat   = null;
	}

	/////////////////////////////////////////////////////////
	//                    Model Accessors
	/////////////////////////////////////////////////////////

	public void addNewHMM(int type, double lambda)
	{
		HMM h = new HMM(type);
		addHMM(h, lambda);
	}

	public void addHMM(HMM h, double lambda)
	{
		_alHMM.add(h);
		_alLambda.add(new Double(lambda));
	}

	public int getNumHMMs()
	{
		return _alHMM.size();
	}

	public HMM getHMM(int index)
	{
		return (HMM)_alHMM.get(index);
	}

	public double getLambda(int index)
	{
		return ((Double)_alLambda.get(index)).doubleValue();
	}

	/////////////////////////////////////////////////////////
	//    Training and Shrinkage-Based-Viterbi Algorithms
	/////////////////////////////////////////////////////////
	public void trainHMMs(DocSet traindata)
	{
		// Set the sufficient statistics for this data
		_ssStat = traindata.getSufficientStats();

		// Normalize the wights to sum to 1.0
		double total = 0.0;
		int    hmms  = getNumHMMs();
		for (int i=0; i<hmms; i++)
			total += getLambda(i);

		//System.out.println("Training...");
		for (int i=0; i<hmms; i++)
		{
			//System.out.println(getLambda(i) + " / " + total
			//				   +  " = " + new Double(getLambda(i)/total));
			_alLambda.set(i, new Double(getLambda(i)/total));
		}

		// Train the individual HMM's
		for (int i=0; i<hmms; i++)
			getHMM(i).train(_ssStat);
	}

	public void extractAllStateModels(DocSet testdata)
	{
		int i, n = testdata.getNumDocuments();
		for (i=0; i<n; i++)
			extractStateModel(testdata, i);
	}

	public void extractStateModel(DocSet testdata, int ndoc)
	{
		// We are given an unlabeled document, we need to
		// find the highest probability path through it and
		// label it along the way...
		int s, s_prime, t, obs;
		int TIME_STEPS = testdata.getDocLength(ndoc); // Includes END
		int STATES     = DocToken._NUM_VALID_STATES;
		int START      = DocToken._START;
		int END        = DocToken._END;
		int ILLEGAL    = DocToken._UNKNOWN;
		double[][] _a2LogPathProb = new double[STATES][TIME_STEPS];
		int[][]    _a2Backtrace   = new int[STATES][TIME_STEPS];
		double new_score;

		// Initialize
		for (t=0; t<TIME_STEPS; t++)
			for (s=0; s<STATES; s++)
			{
				_a2LogPathProb[s][t] = _NINF;
				_a2Backtrace[s][t]   = ILLEGAL;
			}
		_a2LogPathProb[START][0] = Math.log(1.0);

		// Run the algorithm
		for (t=0; t<(TIME_STEPS-1); t++)
			for (s=START; s<=END; s++)
			{
				obs = testdata.getDocumentToken(ndoc, t+1)._nObsIndex;
				for (s_prime=START; s_prime<=END; s_prime++)
				{
					new_score = _a2LogPathProb[s][t] + 
						Math.log(getTransProb(s_prime, s)) +
						Math.log(getObsProb(obs, s_prime));
					
					if ((_a2LogPathProb[s_prime][t+1] == _NINF) ||
						(new_score > _a2LogPathProb[s_prime][t+1]))
					{
						_a2LogPathProb[s_prime][t+1] = new_score;
						_a2Backtrace[s_prime][t+1]   = s;
					}
				}
			}

		// Label the states with the backtrace from
		// the end state
		t = TIME_STEPS;
		s = END;
		while (--t >= 0)
		{
			testdata.getDocumentToken(ndoc, t)._nClass = s;
			s = _a2Backtrace[s][t];
		}   
	}

	/////////////////////////////////////////////////////////
	//            Probability Mixture Functions
	/////////////////////////////////////////////////////////

	// The predictions of all models are mixed by combining
	// the translated probabilities from each sub-HMM.
	// The sub-HMM's take care of mapping the requested states
	// onto their internal states.

	// Compute the mixture transition probabilities
	public double getTransProb(int to_state, int from_state)
	{
		double accum = 0.0;
		int h, nh = getNumHMMs();
		HMM hmm;

		for (h=0; h<nh; h++)
		{
			accum += getLambda(h)*
				(getHMM(h).getTransProb(to_state,from_state));
		}
		
		return accum;
	}

	// Compute the mixture observation probabilities
	public double getObsProb(int obs, int state)
	{
		double accum = 0.0;
		int h, nh = getNumHMMs();
		HMM hmm;

		for (h=0; h<nh; h++)
		{
			accum += getLambda(h)*
				(getHMM(h).getObsProb(obs,state));
		}
		
		return accum;
	}

	//////////////////////////////////////////////////////////////
	//                      Print Utilities
	//////////////////////////////////////////////////////////////
	public void printProbabilities(PrintStream ps)
	{
		int h, nh = getNumHMMs();
		int i, j;
		HMM hmm;
		for (h=0; h<nh; h++)
		{
			ps.println("\nHMM #" + (h+1) + "(" + 
					   _df.format(getLambda(h))
					   + "):\n-------\n");
			hmm = getHMM(h);
			hmm.printProbabilities(ps);
		}

		// Calculate transition probabilities...
		ps.println("\nMix. Transition Probabilities: row(to), col(from)\n" +
				   "-------------------------------------------------");
		ps.println("           STA      BKG      PRE      SUF      TAR      END");
		ps.println("         -------  -------  -------  -------  -------  -------");
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
		{
			ps.print(DocToken.getClassificationName(i) + ": ");
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
		    {
				ps.print(_df.format(getTransProb(i,j)) + "  ");
			}
			ps.println();
		}

		// Calculate observation probabilities
		ps.println("\nMix. Observation Probabilities: row(obs), col(state)\n" +
				   "----------------------------------------------------");
		ps.println("           STA      BKG      PRE      SUF      TAR      END");
		ps.println("         -------  -------  -------  -------  -------  -------");
		for (i=0; i<_ssStat._nWords; i++)
		{
			ps.print(formatWord(i) + ": ");
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
			{
				ps.print(_df.format(getObsProb(i,j)) + "  ");
			}
			ps.println();
		}
		ps.println();
	}

	// Returns an observation word of length 7
	public String formatWord(int index)
	{
		int i, len;
		String ret = _ssStat.getWord(index);
		if (ret.length() > 7)
			ret = ret.substring(0,7);
		len = ret.length();
		for (i=len; i<7; i++)
			ret += " ";
		return ret;
	}

}
