//////////////////////////////////////////////////////////////////////
//
// File:     SuffStatistics.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     4/12/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.util.*;
import java.lang.*;
import java.text.*;

// Main class - Caches overall sufficient statistics from IO
public class SuffStatistics
{
	/** Constants **/
	public final static int _MAX_WORDS = 10000;
	public final static String _OVERFLOW = "[OVERFLOW]";
	public final static DecimalFormat _df = new DecimalFormat("000");

	/** Members **/
	public int        _nWords;           // Number of words currently
										 // stored
	public ArrayList  _alWords;          // List of words as entered
	public HashMap    _hmWordMap;        // Map from words to obs indices
	public int[]      _aStatePriors;     // Count of state occurrences
	public int[]      _aObsPriors;       // Count of observation occurrences
	public int[][]    _a2StateFromState; // Not symmetric, not conditional...
	                                     // conj. stats for [toState][fromState]
	public int[][]    _a2ObsFromState;   // conj. stats for [obs][state]

	/** Methods **/
	public SuffStatistics()
	{
		_nWords = 0;
		_hmWordMap = new HashMap();
		_alWords   = new ArrayList();
		_hmWordMap.put(_OVERFLOW, new Integer(_nWords++));
		_alWords.add(_OVERFLOW);

		_aStatePriors     = new int[DocToken._NUM_VALID_STATES];
		_aObsPriors       = new int[_MAX_WORDS];
		_a2StateFromState = new int[DocToken._NUM_VALID_STATES][DocToken._NUM_VALID_STATES];
		_a2ObsFromState   = new int[_MAX_WORDS][DocToken._NUM_VALID_STATES];

		resetSufficientStats();
	}

	public void resetSufficientStats()
	{
		int i, j;
		
		// Clear the priors
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
			_aStatePriors[i] = 0;

		for (i=0; i<_MAX_WORDS; i++)
			_aObsPriors[i] = 0;

		// Clear the transition probabilities
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				_a2StateFromState[i][j] = 0;
		
		// Clear the observation probabilities
		for (i=0; i<_MAX_WORDS; i++)
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
				_a2ObsFromState[i][j] = 0;
	}

	public void AddStateFromState(int to_state, int from_state)
	{
		if (to_state == DocToken._BOTH)
		{
			if (from_state == DocToken._BOTH)
			{
				// Both states are _BOTH
				_a2StateFromState[DocToken._PREFIX][DocToken._PREFIX]++;
				_a2StateFromState[DocToken._PREFIX][DocToken._SUFFIX]++;
				_a2StateFromState[DocToken._SUFFIX][DocToken._PREFIX]++;
				_a2StateFromState[DocToken._SUFFIX][DocToken._SUFFIX]++;
			}
			else
			{
				// Just to_state is _BOTH
				_a2StateFromState[DocToken._PREFIX][from_state]++;
				_a2StateFromState[DocToken._SUFFIX][from_state]++;
			}
		}
		else
		{
			if (from_state == DocToken._BOTH)
			{
				// Just from_state is _BOTH
				_a2StateFromState[to_state][DocToken._PREFIX]++;
				_a2StateFromState[to_state][DocToken._SUFFIX]++;
			}
			else
			{
				// Neither are _BOTH
				_a2StateFromState[to_state][from_state]++;
			}
		}
	}

	public void AddObsFromState(int obs, int state)
	{
		// Check for _BOTH here... this can get hairy
		// if both the from and to are _BOTH
		if (state == DocToken._BOTH)
		{
			_a2ObsFromState[obs][DocToken._PREFIX]++;
			_a2ObsFromState[obs][DocToken._SUFFIX]++;
		}
		else
			_a2ObsFromState[obs][state]++;
	}

	public void incStatePriors(int state)
	{
		// Check for _BOTH here... this can get hairy
		// if both the from and to are _BOTH
		if (state == DocToken._BOTH)
		{
			_aStatePriors[DocToken._PREFIX]++;
			_aStatePriors[DocToken._SUFFIX]++;
		}
		else
			_aStatePriors[state]++;
	}

	public void incObsPriors(int obs)
	{
		_aObsPriors[obs]++;
	}

	public String getWord(int index)
	{
		return (String)_alWords.get(index);
	}

	// Retrieve a word and add it if not there
	public int getIndex(String word)
	{
		Integer i = (Integer)_hmWordMap.get(word);
		if (i == null) // Add to map
			{
				if (_nWords >= _MAX_WORDS)
				{
					System.err.println("[**Max words reached**]");
					return 0; // The [OVERFLOW] word
				}
				
				// There is still space, so add this word
				_hmWordMap.put(word, new Integer(_nWords));
				_alWords.add(word);
				return _nWords++;
			}
		else
			return i.intValue();
	}

	// Rather than add an unknown word, just returns (-1)
	public int getIndexNoAdd(String word)
	{
		Integer i = (Integer)_hmWordMap.get(word);
		if (i == null) // Add to map
			return (-1);
		else
			return i.intValue();
	}

	// Print out the transition and observation probabilities
	public void printStatistics(PrintStream ps)
	{
		int i, j;
		ps.println("\nTransitions: row(to), col(from)\n" +
				   "----------------------------------");
		ps.println("         STA  BKG  PRE  SUF  TAR  END");
		ps.println("         ---  ---  ---  ---  ---  ---");
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
		{
			ps.print(DocToken.getClassificationName(i) + "  ");
			for (j=0; j<DocToken._NUM_VALID_STATES; j++)
		    {
				ps.print(_df.format(_a2StateFromState[i][j]) + "  ");
			}
			ps.println();
		}

		ps.println("\nBase probabilities: row(to), col(from)\n" +
				   "--------------------------------------");
		for (i=0; i<DocToken._NUM_VALID_STATES; i++)
		{
			ps.println(DocToken.getClassificationName(i) + ": " +
					 _aStatePriors[i]);
		}
		ps.println();

		// Calculate observation probabilities
		//ps.println("\nObservation Probabilities: row(obs), col(state)\n" +
		//		   "-----------------------------------------------");
		//for (i=0; i<_ssStat._nWords; i++)
		//	{
		//	for (j=0; j<DocToken._NUM_VALID_STATES; j++)
		//	{
		//		ps.print(_df.format(_a2Obs[i][j]) + "  ");
		//	}
		//ps.println();
		//}
		//ps.println();
	}

}
