//////////////////////////////////////////////////////////////////////
//
// File:     Template.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     4/12/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.util.*;
import java.lang.*;

// Main class - Represents a template to extract,
//
// Currently this only handles templates consisting of
// a single target type and only handles a single 
// template.
//
// More complex template building could be developed
// from aggregates of multiple templates (e.g. for 
// classes: title, time, loc, instructor) or from
// allowing more complex target models in the HMM
// (rather than just the current single uniform target
//  model).
public class Template
{
	/** Constants **/
	public final static int _MAX_INPUT = 4096;
	public final static int _TREAT_AS_UNLABELED = (-1);

	/** Members **/
   	public int       _nContextWidth; // Width of prefix/suffix	
	public String    _sTargetName;   // Name of target in file comments:
						  	  	     // <!--?-NAME-->
	public boolean   _bIgnoreTags;   // Ignore tags?
	public ArrayList _alType ;       // List of HMM model types
	public ArrayList _alLambda;      // List of HMM lambdas
	public String    _sFilename;     // Filename parsed from

	/** Methods **/
	public Template()
	{
		_nContextWidth = _TREAT_AS_UNLABELED;
		_sTargetName   = "[NONE]";
		_bIgnoreTags   = false;
		_alType        = new ArrayList();
		_alLambda      = new ArrayList();		
	}

	public Template(String filename)
	{
		_nContextWidth = _TREAT_AS_UNLABELED;
		_sTargetName   = "[NONE]";
		_bIgnoreTags   = false;
		_sFilename     = filename;
		_alType        = new ArrayList();
		_alLambda      = new ArrayList();		

        if (filename == null) return;
		
        try {
            String varname, value, value2;
            InputStream in = new FileInputStream(filename);
            byte[] b = new byte[_MAX_INPUT];
            
            int l = in.read(b);
            String s = new String(b);
            StringTokenizer st = new StringTokenizer(s.substring(0,l-1));
            
            while (st.hasMoreTokens())
                {
                    varname = st.nextToken().trim();
                    if (!st.hasMoreTokens()) break;
                    value = st.nextToken().trim();

                    //setBindings(varname, value);
					if (varname.toUpperCase().compareTo("CONTEXT-WIDTH") == 0)
						_nContextWidth = Integer.parseInt(value);
					else if (varname.toUpperCase().compareTo("TARGET") == 0)
						_sTargetName = value;
					else if (varname.toUpperCase().compareTo("IGNORE-TAGS") == 0)
						_bIgnoreTags =
							(value.toUpperCase().compareTo("FALSE") != 0);
					else if (varname.toUpperCase().compareTo("ADD-MODEL") == 0)
						{
							if (!st.hasMoreTokens())
								{
									System.out.println("\nADD-MODEL takes"
													   +" two args, halting...");
									System.exit(1);
								}
							value2 = st.nextToken().trim();
							addModel(value, value2);
						}

                }
        } catch (IOException e) 
		{
			System.out.println("\nError while processing template '" +
							   filename + "', halting...");
			System.exit(1);
		} catch (NumberFormatException e)
		{
			System.out.println("\nMalformed numeric value in template '" +
							   filename + "', halting...");
			System.exit(1);
		}
		
	}	

	public void addModel(String type, String lambda)
	{
		int    ntype   = -1;
		double dlambda = -1.0;

		if (type.toUpperCase().compareTo("UNIFORM") == 0)
			ntype = HMM._UNIFORM;
		else if (type.toUpperCase().compareTo("CONTEXT") == 0)
			ntype = HMM._CONTEXT;
		else if (type.toUpperCase().compareTo("ALL") == 0)
			ntype = HMM._ALL;
		else
		{
			System.out.println("\n'" + type + "' is not a valid HMM option in "
							   + ", halting...");
			System.exit(1);
		}

		try {
			dlambda = Double.parseDouble(lambda);
		} catch (NumberFormatException e)
		{
			System.out.println("\nMalformed HMM lambda in template '" +
							   _sFilename + "', halting...");
			System.exit(1);
		}
			
		_alType.add(new Integer(ntype));
		_alLambda.add(new Double(dlambda));
	}

	public int getNumModels()
	{
		return _alType.size();
	}

	public int getType(int index)
	{
		return ((Integer)_alType.get(index)).intValue();
	}

	public double getLambda(int index)
	{
		return ((Double)_alLambda.get(index)).doubleValue();
	}

	public void printTemplateInfo(PrintStream ps)
	{
		ps.println("\nUsing template '" + _sFilename +
				   "' with the following options:");
		ps.println("==============================="+
				   "===============================\n");

		ps.println("  Context Width: " + _nContextWidth);
		ps.println("  Target Name:   " + _sTargetName);
		ps.println("  Ignore Tags:   " + 
				   (_bIgnoreTags ? "Yes" : "No"));
		for (int i=0; i<getNumModels(); i++)
			ps.println("  HMM #"+(i+1)+":        " + 
					   printType(getType(i)) + "/" + 
					   getLambda(i));
		ps.println();
	}

	public String printType(int type)
	{
		switch (type)
		{
		case HMM._UNIFORM: return "Uniform";
		case HMM._CONTEXT: return "Context";
		case HMM._ALL:     return "All";
		}   
		return "[ILLEGAL]";
	}

	public Template cloneUnlabeled()
	{
		Template t = new Template();
		
		t._nContextWidth = _TREAT_AS_UNLABELED;
		t._sTargetName   = _sTargetName;
		t._bIgnoreTags   = _bIgnoreTags;

		return t;
	}
}
