//////////////////////////////////////////////////////////////////////
//
// File:     TemplateExtractor.java
// Project:  Java-based html tag extractor
// Author:   Scott Sanner, Stanford University
// Date:     4/12/2000
//
//////////////////////////////////////////////////////////////////////

// Package definition
package Extractor;

import java.io.*;
import java.util.*;
import java.lang.*;

// Main class - Main Interface for Template Extractor
public class TemplateExtractor
{
	/** Constants **/

	/** Members **/
	public DocSet   _dsTrain;
	public DocSet   _dsTest;
	public DocSet   _dsAnswers;
	public Template _tTemplate;
	public boolean  _bCompare;

	public SuffStatistics _ssStat;
	public MixtureHMM     _mhmmModel;

	/** Methods **/
	public TemplateExtractor(String trainset, String testset, 
							 String template, boolean bcomp)
	{
		// Load template
		_tTemplate = new Template(template);
		_ssStat    = new SuffStatistics();
		DocToken.setSuffStats(_ssStat); // Given the tokenizer
		                                // access to sstats so that
		                                // it can index words!

		// Load the documents
		_bCompare  = bcomp;
		_dsTrain   = new DocSet(trainset, _tTemplate, _ssStat);
		_dsTest    = new DocSet(testset,  _tTemplate.cloneUnlabeled(), null);
		_dsAnswers = (_bCompare) ? new DocSet(testset,  _tTemplate, null) : null;

		// Print out the templates and documents
		_tTemplate.printTemplateInfo(System.out);
		//printDocSets(System.out);		
	}

	public void extract(PrintStream ps)
	{
		ps.println("\nTraining and extracting state sequence...\n");

		// Initialize HMM
		initModel(_dsTrain, _tTemplate);

		// Run the HMM Algorithms to label the states
		labelStates(_dsTest);
		//_dsTest.printAllDocuments(ps);
		//ps.println("\n");

		// Export the extracted data
		ps.println("\nTest Data Export:\n==================\n");
		exportTemplateData(_dsTest, ps);
		if (_bCompare)
		{
			ps.println("\nAnswer Data Export:\n====================\n");
			exportTemplateData(_dsAnswers, ps);
		}

		// Calculate Results
		if (_bCompare)
			_dsAnswers.printComparisonResults(_dsTest, ps);
	}

	public void initModel(DocSet train, Template temp)
	{
		// Add each specified HMM to a new mixture model
		_mhmmModel = new MixtureHMM();
		for (int i=0; i<temp.getNumModels(); i++)
			_mhmmModel.addNewHMM(temp.getType(i), temp.getLambda(i));
		_mhmmModel.trainHMMs(train);
		//_ssStat.printStatistics(System.out);
		//_mhmmModel.printProbabilities(System.out);
	}

	public void labelStates(DocSet test)
	{
		// Run the HMM Algorithms to label the states
		_mhmmModel.extractAllStateModels(test);
	}

	public void exportTemplateData(DocSet test, PrintStream ps)
	{
		// Export the extracted data
		test.exportResults(ps);
	}

	public void printDocSets(PrintStream ps)
	{
		System.out.println("Training Set:\n============\n");
		_dsTrain.printAllDocuments(ps);
		System.out.println("\n");
		
		//System.out.println("Test Set:\n============\n");
		//_dsTest.printAllDocuments(ps);
		//System.out.println("\n");
	}

	public static void main(String[] argv)
	{
		if (argv.length < 4)
		{
			System.out.println("\nUsage: java Extractor.TemplateExtractor\n" +
							   "<trainset> <testset> <template> {comp|nocomp}\n");
			System.exit(1);
		}

		boolean comp = (argv[3].trim().toUpperCase().compareTo("COMP") == 0);
		TemplateExtractor te = 
			new TemplateExtractor(argv[0], argv[1], argv[2], comp);

		te.extract(System.out);
	}
}
